//
// Created by koerberm on 28.02.19.
//

#ifndef EPHSA_EVENT_H
#define EPHSA_EVENT_H

#define NUM_VALUES 6
#define NUM_PRED_VALUES 6

#ifdef CPU_COMPILE
    #include <cmath>
    #include <random>
    #include <iostream>
    using namespace std;
#else
    #include "/opt/rocm/include/ockl_hsa.h"
#endif

typedef struct {
    unsigned long timestamp;
    float values[NUM_VALUES];
} event;

typedef struct {
    unsigned long timestamp;
    float values[2*NUM_VALUES];
} joined_event;

typedef struct {
    unsigned long timestamp;
    float values[1];
} aggregated_event;

inline float agg_init( const event *in ) {
    return in->values[0];
}

inline void agg_merge( float *accum, const float value) {
//    *accum = fmax(*accum,value);
    *accum += value;
}

#ifndef CPU_COMPILE
inline float agg_wf_merge( float value ) {
//    return __ockl_wfred_max_f32(value);
    return __ockl_wfred_add_f32(value);
}
#endif

inline aggregated_event agg_finish( const event *e, float accum ) {
    aggregated_event result;
    result.timestamp = e->timestamp;
    result.values[0] = accum;
    return result;
}

inline bool filter_predicate( const event *e ) {
    float d_origin = 0;
    for ( uint i = 0; i < NUM_PRED_VALUES; i++ ) {
        d_origin += (e->values[i]*e->values[i]);
    }
    return sqrt(d_origin) <= 1.0f;
}

inline bool join_predicate( const event *el, const event *er ) {
//    return fabs(el->values[0]-er->values[0]) <= 1;
//    return false;
    float d = 0;
    for ( uint i = 0; i < NUM_PRED_VALUES; i++ ) {
        d += fabs(el->values[i]*el->values[i] - er->values[i]*er->values[i]);
    }

    return sqrt(d) <= 1.0f;
}

inline joined_event compose( const event *el, const event *er ) {
    const unsigned long ts = el->timestamp > er->timestamp ? el->timestamp : er->timestamp;
//    joined_event r = {ts,el->value,er->value};
    joined_event r;
    r.timestamp = ts;
    for ( uint i = 0; i < NUM_VALUES; i++ ) {
        r.values[i] = el->values[i];
        r.values[i+NUM_VALUES] = er->values[i];
    }
    return r;
}


#ifdef CPU_COMPILE
inline event create_event(uint64_t ts, std::uniform_real_distribution<float> &dis, std::mt19937 &gen ) {
    event result;
    result.timestamp = ts;
    for ( uint i = 0; i < NUM_VALUES; i++ ) {
        result.values[i] = (i==0) ? dis(gen) : 0.0f;
    }
    return result;
}

inline std::ostream& operator<<(std::ostream &o, const event& p) {
    o << "(" << p.timestamp;
    for ( uint i = 0; i < NUM_VALUES; i++ ) {
        o << ", " << p.values[i];
    }
    return o << ")";
}

inline std::ostream& operator<<(std::ostream &o, const joined_event& p) {
    o << "(" << p.timestamp;
    for ( uint i = 0; i < 2*NUM_VALUES; i++ ) {
        o << ", " << p.values[i];
    }
    return o << ")";
}

inline std::ostream& operator<<(std::ostream &o, const aggregated_event& p) {
    o << "(" << p.timestamp;
    for ( uint i = 0; i < 2*NUM_VALUES; i++ ) {
        o << ", " << p.values[i];
    }
    return o << ")";
}
#endif

#endif //EPHSA_EVENT_H
