//
// Created by koerberm on 28.02.19.
//

#ifndef EPHSA_EVENT_LATENCY_H
#define EPHSA_EVENT_LATENCY_H

#define L_NUM_VALUES 6

#define L_NUM_PRED_VALUES 6

#ifdef CPU_COMPILE
    #include <cmath>
    #include <random>
    #include <iostream>
    #include <cstring>
    using namespace std;
#endif

typedef struct event_latency_t {
    unsigned long timestamp;
    unsigned long in_time;
    unsigned long out_time;
    float values[L_NUM_VALUES];
} event_latency;

typedef struct {
    unsigned long timestamp;
    unsigned long in_time;
    unsigned long out_time;
    float values[2*L_NUM_VALUES];
} joined_event_latency;

inline bool filter_predicate( const event_latency *e ) {
    float d_origin = 0;
    for ( uint i = 0; i < L_NUM_PRED_VALUES; i++ ) {
        d_origin += (e->values[i]*e->values[i]);
    }
    return sqrt(d_origin) <= 1.0f;
}

inline bool join_predicate( const event_latency *el, const event_latency *er ) {
    float d = 0;
    for ( uint i = 0; i < L_NUM_PRED_VALUES; i++ ) {
        d += fabs(el->values[i]*el->values[i] - er->values[i]*er->values[i]);
    }

    return sqrt(d) <= 1.0f;
}

inline joined_event_latency compose( const event_latency *el, const event_latency *er ) {
    const unsigned long ts = el->timestamp > er->timestamp ? el->timestamp : er->timestamp;
//    joined_event r = {ts,el->value,er->value};
    joined_event_latency r;
    r.timestamp = ts;
    r.in_time = max(el->in_time,er->in_time);
    r.out_time = 0;
    for ( uint i = 0; i < L_NUM_VALUES; i++ ) {
        r.values[i] = el->values[i];
        r.values[i+L_NUM_VALUES] = er->values[i];
    }
    return r;
}

#ifdef CPU_COMPILE
inline event_latency create_event_latency(uint64_t ts, std::uniform_real_distribution<float> &dis, std::mt19937 &gen ) {
    event_latency result;
    result.timestamp = ts;
    result.in_time = 0;
    result.out_time = 0;
    for ( uint i = 0; i < L_NUM_VALUES; i++ ) {
        result.values[i] = (i==0) ? dis(gen) : 0.0f;
    }
    return result;
}

inline std::ostream& operator<<(std::ostream &o, const event_latency& p) {
    o << "(" << p.timestamp << ", " << p.in_time << ", " << p.out_time;
    for ( uint i = 0; i < L_NUM_VALUES; i++ ) {
        o << ", " << p.values[i];
    }
    return o << ")";
}

inline std::ostream& operator<<(std::ostream &o, const joined_event_latency& p) {
    o << "(" << p.timestamp << ", " << p.in_time << ", " << p.out_time;
    for ( uint i = 0; i < 2*L_NUM_VALUES; i++ ) {
        o << ", " << p.values[i];
    }
    return o << ")";
}
#endif

#endif //EPHSA_EVENT_H
