//
// Created by koerberm on 07.03.19.
//

#include "latency_addons.h"
#include "event_latency.h"

namespace ep {

    template<class T>
    event_queue_latency<T>::event_queue_latency(uint32_t capacity, const hsa::hsa_platform &platform,
                                             const hsa::hsa_agent &agent, bool set_out)
        : event_queue<T>(capacity,platform,agent), set_out(set_out), platform(platform) {

    }

    template<class T>
    hsa_signal_value_t event_queue_latency<T>::add(const T &event) {
        return event_queue<T>::add(event);
    }

    template<>
    hsa_signal_value_t event_queue_latency<event_latency>::add(const event_latency &event) {
        wait_not_full();
        auto addr = buffer->address<event_latency>();
        addr[writer_w_idx & mask] = event;

        if (set_out) {
            addr[writer_w_idx & mask].out_time = platform.get_timestamp();
        }
        else {
            addr[writer_w_idx & mask].in_time = /*std::chrono::steady_clock::now();//*/platform.get_timestamp();
        }

        w_signal->add_screlease(1);
        return writer_w_idx++;
    }

    template<>
    hsa_signal_value_t event_queue_latency<joined_event_latency>::add(const joined_event_latency &event) {
        wait_not_full();
        auto addr = buffer->address<joined_event_latency>();
        addr[writer_w_idx & mask] = event;

        if (set_out) {
            addr[writer_w_idx & mask].out_time = platform.get_timestamp();
        }
        else {
            addr[writer_w_idx & mask].in_time = platform.get_timestamp();
        }

        w_signal->add_screlease(1);
        return writer_w_idx++;
    }


    std::ostream &operator<<(std::ostream &s, const ep::latency_stats &p) {
        return s << "<" << p.min << "," << p.max << "," << (p.count > 0 ? (p.sum/p.count) : 0lu) << ">";
    }

    /*
     * Latency measurement
     */

    latency_stats::latency_stats() : min(UINT64_MAX), max(0), sum(0), count(0) {
    }

    template<class T>
    latency_measure<T>::latency_measure(const shared_ptr<event_queue<T>> &in_queue) :
        event_sink_cpu<T>(in_queue,1,in_queue->capacity) {
    }

    template<class T>
    latency_stats latency_measure<T>::get_stats() {
        return stats;
    }

    template<class T>
    void latency_measure<T>::process_element(const T &event) {
        (void) event;
    }

    template<>
    void latency_measure<event_latency>::process_element(const event_latency &event) {
        uint64_t l = event.out_time-event.in_time;
        if ( stats.sum > stats.sum+l )
            throw std::exception();
        stats.sum += l;
        stats.count++;
        stats.min = std::min(stats.min,l);
        stats.max = std::max(stats.max,l);
    }

    template<>
    void latency_measure<joined_event_latency>::process_element(const joined_event_latency &event) {
        uint64_t l = event.out_time-event.in_time;
        stats.sum += l;
        stats.count++;
        stats.min = std::min(stats.min,l);
        stats.max = std::min(stats.max,l);
    }

    template class event_queue_latency<event_latency>;
    template class event_queue_latency<joined_event_latency>;

    template class latency_measure<event_latency>;
    template class latency_measure<joined_event_latency>;

}