//
// Created by koerberm on 07.03.19.
//

#ifndef EPHSA_LATENCY_ADDONS_H
#define EPHSA_LATENCY_ADDONS_H

#include <thread>
#include "event_latency.h"
#include "../event_queue.h"
#include "../cpu_operators.h"

namespace ep {

    template<class T>
    class event_queue_latency : public event_queue<T> {
    public:
        event_queue_latency(uint32_t capacity, const hsa::hsa_platform &platform, const hsa::hsa_agent &agent,
                            bool set_out);

        hsa_signal_value_t add(const T &event);

    private:
        const bool set_out;
        const hsa::hsa_platform &platform;
    };


    class latency_stats {
    public:
        latency_stats();

        uint64_t min;
        uint64_t max;
        uint64_t sum;
        uint64_t count;

        friend std::ostream& operator<<(std::ostream& s, const latency_stats& p);
    };

    /*
     * Measures the latency of the given event from timestamp to receipt in output queue
     */
    template<class T>
    class latency_measure : public event_sink_cpu<T> {
    public:
        explicit latency_measure(const std::shared_ptr<event_queue<T>> &in_queue);
        latency_stats get_stats();
    protected:
        void process_element( const T &event);
    private:
        latency_stats stats;
    };
}
#endif //EPHSA_EVENT_QUEUE_LATENCY_H
