//
// Created by koerberm on 09.03.18.
//

#ifndef EPHSA_STREAM_OPERATOR_H
#define EPHSA_STREAM_OPERATOR_H


#include <cstdint>
#include "event_queue.h"
#include "../rocm/shared.h"

namespace ep {

    template <class T>
    class count_window {
    public:
        count_window(uint64_t capacity, const hsa::hsa_agent &agent);

        count_window(const count_window<T> &copy);
        count_window(count_window<T> &&mov) noexcept;

        count_window<T> &operator=(const count_window<T> &);
        count_window<T> &operator=(count_window<T> &&) noexcept;

        ~count_window();
        void add( const T &event );
        window_def to_struct() const;
        uint64_t get_capacity() const { return capacity; };
        uint64_t get_inserts() const { return inserts; };
        const T* get_base() const {return buffer->address<T>();};
        template <class E>
        friend std::ostream& operator<<(std::ostream& s, const count_window<E>& w);
    private:
        std::unique_ptr<hsa::hsa_buffer> buffer;
        uint64_t capacity;
        uint64_t inserts;
    };

    template<class T>
    std::ostream &operator<<(std::ostream &s, const count_window<T> &w) {
        s << "Count window (" << w.inserts << " inserts): ";
        for ( uint64_t i = 0; i < std::min(w.inserts, w.capacity); i++ ) {
            s << w.get_base()[i] << ", ";
        }
        return s;
    }

    /*
     * IFace Stuff
     */

    template<class T>
    class event_sink {
    public:
        explicit event_sink(const std::shared_ptr<event_queue<T>> &queue);
        virtual ~event_sink();
        event_queue<T> & get_in_queue() { return *in_queue; }

        virtual uint64_t get_processing_latency() = 0;
        virtual void stop() = 0;
    protected:
        const std::shared_ptr<event_queue<T>> in_queue;
    };

    template<class T1, class T2>
    class binary_event_sink {
    public:
        explicit binary_event_sink(const std::shared_ptr<event_queue<T1>> &queue1, const std::shared_ptr<event_queue<T2>> &queue2);
        virtual ~binary_event_sink();

        event_queue<T1> & get_in_queue1() const { return *in_queue1; }
        event_queue<T2> & get_in_queue2() const { return *in_queue2; }

        virtual uint64_t get_processing_latency() = 0;
        virtual void stop() = 0;
    protected:
        const std::shared_ptr<event_queue<T1>> in_queue1;
        const std::shared_ptr<event_queue<T2>> in_queue2;
    };

    template<class T>
    class event_source {
    public:
        explicit event_source(const std::shared_ptr<event_queue<T>> &out_queue);
        virtual ~event_source();
        event_queue<T> & get_out_queue() { return *out_queue; }
    protected:
        const std::shared_ptr<event_queue<T>> out_queue;
    };
}

#endif //EPHSA_STREAM_OPERATOR_H
