//
// Created by koerberm on 06.03.18.
//

#include <fstream>

#include "hsa_platform_programs.h"
#include "hsa_platform_util.h"
#include "hsa_platform_core.h"

#define check(status) \
if ((status) != HSA_STATUS_SUCCESS) { \
    throw hsa::hsa_exception(status); \
}

namespace hsa {
    /*
     * Utilities
     */
    const char *map_linkage(const hsa_symbol_linkage_t &l) {
        switch (l) {
            case HSA_SYMBOL_LINKAGE_PROGRAM: return "program";
            case HSA_SYMBOL_LINKAGE_MODULE: return "module";
            default: throw hsa_exception("Unknown symbol linkage.");
        }
    }

    const char *map_segment(const hsa_variable_segment_t &l) {
        switch (l) {
            case HSA_VARIABLE_SEGMENT_GLOBAL: return "global";
            case HSA_VARIABLE_SEGMENT_READONLY: return "read-only";
            default: throw hsa_exception("Unknown symbol linkage.");
        }
    }

    const char *map_allocation(const hsa_variable_allocation_t &l) {
        switch (l) {
            case HSA_VARIABLE_ALLOCATION_AGENT: return "agent";
            case HSA_VARIABLE_ALLOCATION_PROGRAM: return "program";
            default: throw hsa_exception("Unknown symbol linkage.");
        }
    }

//        No finalization in ROCm 2.x
//    std::unique_ptr<hsa_module> module_from_file(const char *file_name) {
//
//        std::ifstream in(file_name, std::ios::in | std::ios::binary);
//        if (in)
//        {
//            hsa_ext_module_t module = nullptr;
//            in.seekg(0, std::ios::end);
//            long t_res = in.tellg();
//            if ( t_res < 0 )
//                throw hsa_exception("Could not determine size of HSAIL module file.");
//
//            auto size = (unsigned long) (t_res);
//            module = (hsa_ext_module_t) malloc(size);
//            memset(module,0,size);
//            in.seekg(0, std::ios::beg);
//            in.read((char*) module, size);
//            in.close();
//            return std::make_unique<hsa_module>( module );
//        }
//        throw hsa_exception("Could not open HSAIL module file.");
//    }

    std::shared_ptr<hsa_code_object> code_object_from_file(const std::string &filename, const hsa::hsa_agent &agent) {
        auto full_file_name = filename + "." + agent.name;
        std::ifstream in(full_file_name.c_str(), std::ios::in | std::ios::binary);
        if (in)
        {
            in.seekg(0, std::ios::end);
            long t_res = in.tellg();
            if ( t_res < 0 )
                throw hsa_exception("Could not determine size of code object file at: " + full_file_name);

            auto serialized_code_object_size = (size_t) (t_res);
            void *raw = malloc(serialized_code_object_size);
            if (nullptr == raw)
                throw hsa_exception("Could not allocate memory for code object.");

            memset(raw,0,serialized_code_object_size);
            in.seekg(0, std::ios::beg);
            in.read((char*) raw, serialized_code_object_size);
            in.close();

            return std::make_shared<hsa_code_object>(raw, serialized_code_object_size);
        }
        throw hsa_exception("Could not open code object file at: " + full_file_name);
    }

    /*
     * Finalizer
     */
//  No finalization in ROCm 2.0
//    hsa_finalizer::hsa_finalizer() {
//        hsa_status_t stat;
//        stat = hsa_system_get_extension_table(HSA_EXTENSION_FINALIZER, HSA_FINALIZER_VERSION_MAJOR, HSA_FINALIZER_VERSION_MINOR, &table);
//        check(stat);
//    }
//
//    std::unique_ptr<hsa_program> hsa_finalizer::create_program(const hsa_agent &agent) const {
//        return create_program(agent.machine_model,agent.profile,agent.default_float_rounding_mode);
//    }
//
//    std::unique_ptr<hsa_program> hsa_finalizer::create_program(const hsa_machine_model_t &machine_model,
//                                                               const hsa_profile_t &profile,
//                                                               const hsa_default_float_rounding_mode_t &rounding_mode) const {
//        hsa_status_t stat;
//        hsa_ext_program_t program = {0};
//        stat = hsa_ext_program_create(machine_model,profile,rounding_mode, nullptr,&program);
//        check(stat);
//        return std::make_unique<hsa_program>(program);
//    }
//
//    /*
//     * Program
//     */
//    hsa_program::hsa_program(hsa_ext_program_t program) : handle(program) {
//        hsa_status_t stat;
//        stat = hsa_ext_program_get_info(handle,HSA_EXT_PROGRAM_INFO_PROFILE,&profile); check(stat);
//        stat = hsa_ext_program_get_info(handle,HSA_EXT_PROGRAM_INFO_MACHINE_MODEL,&machine_model); check(stat);
//        stat = hsa_ext_program_get_info(handle,HSA_EXT_PROGRAM_INFO_DEFAULT_FLOAT_ROUNDING_MODE,&rounding_mode); check(stat);
//        stat = hsa_system_get_extension_table(HSA_EXTENSION_FINALIZER,HSA_FINALIZER_VERSION_MAJOR,HSA_FINALIZER_VERSION_MINOR,&table); check(stat);
//    }
//
//    hsa_program::~hsa_program() {
//        hsa_ext_program_destroy(handle);
//    }
//
//    void hsa_program::add_module(std::shared_ptr<hsa_module> module) {
//        hsa_status_t stat;
//
//        stat = table.hsa_ext_program_add_module(handle,module->blob);
//        if ( HSA_STATUS_SUCCESS == stat )
//            modules.emplace_back(module);
//        check(stat);
//    }
//
//    std::unique_ptr<hsa_code_object> hsa_program::finalize(const hsa_agent &agent) {
//        hsa_code_object_t code_object = {0};
//        hsa_status_t stat;
//
//        hsa_ext_control_directives_t control_directives;
//        ::memset(&control_directives, 0, sizeof(hsa_ext_control_directives_t));
//
//
//
//        stat = table.hsa_ext_program_finalize(handle, agent.isa_info.handle, 0,control_directives,
//                                              nullptr, HSA_CODE_OBJECT_TYPE_PROGRAM, &code_object);
//
//        check(stat);
//
//        return std::make_unique<hsa_code_object>(code_object);
//    }
//
//    /*
//     * Module
//     */
//    hsa_module::hsa_module(hsa_ext_module_t blob) : blob(blob) {
//
//    }
//
//    hsa_module::~hsa_module() {
//        free(blob);
//    }


    /*
     * Code Object
     */
    hsa_code_object::~hsa_code_object() {
        hsa_code_object_reader_destroy(reader);
        free(data);
    }

    hsa_code_object::hsa_code_object(void *data, size_t size) : data(data), data_size(size) {
        hsa_status_t stat;
        stat = hsa_code_object_reader_create_from_memory(data,size,&reader); check(stat);
    }


    /*
     * Executable
     */

    exe_symbol::exe_symbol(hsa_executable_symbol_t symbol) : handle(symbol) {
        name = extract_name(HSA_EXECUTABLE_SYMBOL_INFO_NAME_LENGTH, HSA_EXECUTABLE_SYMBOL_INFO_NAME);
        module_name = extract_name(HSA_EXECUTABLE_SYMBOL_INFO_MODULE_NAME_LENGTH, HSA_EXECUTABLE_SYMBOL_INFO_MODULE_NAME);
        hsa_status_t stat = hsa_executable_symbol_get_info(handle, HSA_EXECUTABLE_SYMBOL_INFO_LINKAGE, &linkage);
        check(stat);
        stat = hsa_executable_symbol_get_info(handle, HSA_EXECUTABLE_SYMBOL_INFO_IS_DEFINITION, &is_definition);
        check(stat);
        stat = hsa_executable_symbol_get_info(handle, HSA_EXECUTABLE_SYMBOL_INFO_AGENT, &agent);
        check(stat);
    }

    std::string exe_symbol::extract_name(hsa_executable_symbol_info_t len_prop,
                                         hsa_executable_symbol_info_t val_prop) {
        hsa_status_t stat;
        uint32_t l;
        std::string res;

        stat = hsa_executable_symbol_get_info(handle, len_prop, &l);
        check(stat);
        res.resize(l);
        stat = hsa_executable_symbol_get_info(handle, val_prop, &res[0]);
        check(stat);
        return res;
    }

    exe_symbol_variable::exe_symbol_variable(hsa_executable_symbol_t symbol) : exe_symbol(symbol) {
        hsa_status_t stat;
        stat = hsa_executable_symbol_get_info(handle, HSA_EXECUTABLE_SYMBOL_INFO_VARIABLE_ADDRESS, &address);
        check(stat);
        stat = hsa_executable_symbol_get_info(handle, HSA_EXECUTABLE_SYMBOL_INFO_VARIABLE_ALIGNMENT, &alignment);
        check(stat);
        stat = hsa_executable_symbol_get_info(handle, HSA_EXECUTABLE_SYMBOL_INFO_VARIABLE_ALLOCATION, &allocation);
        check(stat);
        stat = hsa_executable_symbol_get_info(handle, HSA_EXECUTABLE_SYMBOL_INFO_VARIABLE_IS_CONST, &is_const);
        check(stat);
        stat = hsa_executable_symbol_get_info(handle, HSA_EXECUTABLE_SYMBOL_INFO_VARIABLE_SEGMENT, &segment);
        check(stat);
        stat = hsa_executable_symbol_get_info(handle, HSA_EXECUTABLE_SYMBOL_INFO_VARIABLE_SIZE, &size);
        check(stat);
    }

    std::ostream &operator<<(std::ostream &s, const exe_symbol_variable &a) {
        s << "Variable [name: " << a.name << ", module name: " << a.module_name;
        s << ", linkage: " << map_linkage(a.linkage) << ", is definition: " << a.is_definition;
        if ( a.allocation == HSA_VARIABLE_ALLOCATION_AGENT )
            s << ", agent: " << a.agent.handle;
        s << ", address: " << a.address;
        s << ", allocation: " << map_allocation(a.allocation) << ", segment: " << map_segment(a.segment);
        s << ", alignment: " << a.alignment << ", size: " << a.size << ", is const: " << a.is_const << "]";
        return s;
    }

    exe_symbol_kernel::exe_symbol_kernel(hsa_executable_symbol_t symbol) : exe_symbol(symbol) {
        hsa_status_t stat;
        stat = hsa_executable_symbol_get_info(handle, HSA_EXECUTABLE_SYMBOL_INFO_KERNEL_OBJECT,
                                                           &kernel_object);
        check(stat);
        stat = hsa_executable_symbol_get_info(handle, HSA_EXECUTABLE_SYMBOL_INFO_KERNEL_KERNARG_SEGMENT_ALIGNMENT,
                                                     &kernarg_segment_alignment);
        check(stat);
        stat = hsa_executable_symbol_get_info(handle, HSA_EXECUTABLE_SYMBOL_INFO_KERNEL_KERNARG_SEGMENT_SIZE,
                                        &kernarg_segment_size);
        check(stat);
        stat = hsa_executable_symbol_get_info(handle, HSA_EXECUTABLE_SYMBOL_INFO_KERNEL_DYNAMIC_CALLSTACK, &dynamic_callstack);
        check(stat);
        stat = hsa_executable_symbol_get_info(handle, HSA_EXECUTABLE_SYMBOL_INFO_KERNEL_GROUP_SEGMENT_SIZE, &group_segment_size);
        check(stat);
        stat = hsa_executable_symbol_get_info(handle, HSA_EXECUTABLE_SYMBOL_INFO_KERNEL_PRIVATE_SEGMENT_SIZE,
                                        &private_segment_size);
        check(stat);
    }

    std::ostream &operator<<(std::ostream &s, const exe_symbol_kernel &a) {
        s << "Kernel [name: " << a.name << ", module name: " << a.module_name;
        s << ", linkage: " << map_linkage(a.linkage) << ", is definition: " << a.is_definition;
        s << ", agent: " << a.agent.handle << ", kernel object: " << a.kernel_object;
        s << ", kernarg segment size: " << a.kernarg_segment_size;
        s << ", kernarg segment alignment: " << a.kernarg_segment_alignment;
        s << ", group size: " << a.group_segment_size;
        s << ", private size: " << a.private_segment_size;
        s << ", dynamic callstack: " << a.dynamic_callstack << "]";

        return s;
    }

    exe_symbol_indirect_function::exe_symbol_indirect_function(hsa_executable_symbol_t symbol)  : exe_symbol(symbol) {
        hsa_status_t stat;
        stat = hsa_executable_symbol_get_info(handle, HSA_EXECUTABLE_SYMBOL_INFO_INDIRECT_FUNCTION_OBJECT,
                                              &function_object);
        check(stat);

        stat = hsa_executable_symbol_get_info(handle, HSA_EXECUTABLE_SYMBOL_INFO_INDIRECT_FUNCTION_CALL_CONVENTION,
                                                     &call_convention);
        check(stat);
    }

    std::ostream &operator<<(std::ostream &s, const exe_symbol_indirect_function &a) {
        s << "Indirect Function [name: " << a.name << ", module name: " << a.module_name;
        s << ", linkage: " << map_linkage(a.linkage) << ", is definition: " << a.is_definition;
        s << ", agent: " << a.agent.handle << ", function object: " << a.function_object;
        s << ", call convention: " << a.call_convention << "]";
        return s;
    }




    hsa_executable::hsa_executable(hsa_profile_t profile, hsa_default_float_rounding_mode_t rounding_mode) : profile(profile), frozen(false) {
        hsa_status_t stat = hsa_executable_create_alt( profile, rounding_mode, nullptr, &handle);
        check(stat);
    }

    hsa_executable::~hsa_executable() {
        hsa_executable_destroy(handle);
    }

    void hsa_executable::define_global_variable(const std::string &name, void *address) {
        if ( frozen || !code_objects.empty() )
            throw hsa_exception("Cannot define global variable on frozen executable.");
        hsa_status_t stat = hsa_executable_global_variable_define(handle, name.c_str(), address);
        check(stat);
    }

    void hsa_executable::define_agent_global_variable(const std::string &name, void *address, const hsa_agent &agent) {
        if ( frozen || !code_objects.empty() )
            throw hsa_exception("Cannot define agent global variable on frozen executable.");
        hsa_status_t stat = hsa_executable_agent_global_variable_define(handle, agent.handle, name.c_str(), address);
        check(stat);
    }

    void hsa_executable::define_readonly_variable(const std::string &name, void *address, const hsa_agent &agent) {
        if ( frozen || !code_objects.empty() )
            throw hsa_exception("Cannot define read only variable on frozen executable.");
        hsa_status_t stat = hsa_executable_readonly_variable_define(handle, agent.handle, name.c_str(), address);
        check(stat);
    }

    void hsa_executable::load_code_object(std::shared_ptr<hsa_code_object> co, const hsa_agent &agent) {
        if ( frozen )
            throw hsa_exception("Cannot load code object on frozen executable.");

        hsa_loaded_code_object_t lco;
        hsa_code_object_reader_t reader = co->reader;
        hsa_status_t stat = hsa_executable_load_agent_code_object(handle,agent.handle, reader, nullptr, &lco);
        check(stat);
        code_objects.emplace_back( co );
    }

    void hsa_executable::freeze() {
        if ( frozen )
            return;
        frozen = true;

        hsa_status_t stat;
        stat = hsa_executable_freeze(handle, nullptr);
        check(stat);

        stat = hsa_executable_iterate_symbols(handle, symbol_callback<exe_symbol_variable,HSA_SYMBOL_KIND_VARIABLE>, &variable_symbols);
        check(stat);
        stat = hsa_executable_iterate_symbols(handle, symbol_callback<exe_symbol_kernel,HSA_SYMBOL_KIND_KERNEL>, &kernel_symbols);
        check(stat);
        stat = hsa_executable_iterate_symbols(handle, symbol_callback<exe_symbol_indirect_function,HSA_SYMBOL_KIND_INDIRECT_FUNCTION>, &indirect_function_symbols);
        check(stat);
    }

    template<class T, hsa_symbol_kind_t KIND>
    hsa_status_t hsa_executable::symbol_callback(hsa_executable_t exe, hsa_executable_symbol_t symbol, void *data) {
        (void) exe;
        auto v = (std::vector<T> *) data;

        hsa_status_t stat;
        hsa_symbol_kind_t kind;

        stat = hsa_executable_symbol_get_info(symbol, HSA_EXECUTABLE_SYMBOL_INFO_TYPE, &kind);
        check(stat);

        if (kind == KIND)
            v->emplace_back(T(symbol));
        return HSA_STATUS_SUCCESS;
    }

    std::ostream &operator<<(std::ostream &s, const hsa_executable &a) {
        s << "HSA Executable:" << std::endl;
        s << "  Profile: " << map_profile(a.profile) << std::endl;
        s << "  Frozen : " << a.frozen << std::endl;
        if ( a.frozen ) {
            s << "  Kernel Symbols (" << a.kernel_symbols.size() << "):" << std::endl;
            for ( auto &ks : a.kernel_symbols )
                s << "    " << ks << std::endl;
            s << "  Variable Symbols (" << a.variable_symbols.size() << "):" << std::endl;
            for ( auto &vs : a.variable_symbols )
                s << "    " << vs << std::endl;

            s << "  Indirect Function Symbols (" << a.indirect_function_symbols.size() << "):" << std::endl;
            for ( auto &ifs : a.indirect_function_symbols )
                s << "    " << ifs << std::endl;
            return s;
        }
        else {
            s << "  Symbols only available when frozen." << std::endl;
        }

        return s;
    }

    const std::vector<exe_symbol_variable> &hsa_executable::get_variable_symbols() const {
        return variable_symbols;
    }

    const std::vector<exe_symbol_kernel> &hsa_executable::get_kernel_symbols() const {
        return kernel_symbols;
    }

    const std::vector<exe_symbol_indirect_function> &hsa_executable::get_indirect_function_symbols() const {
        return indirect_function_symbols;
    }

    exe_symbol_kernel hsa_executable::kernel_by_name(const std::string &name) const {
        for ( auto &i : kernel_symbols ) {
            if ( i.name.compare(name) == 0 )
                return i;
        }
        throw hsa_exception("No kernel found for this name!");
    }

    hsa_kernel_dispatch_packet::hsa_kernel_dispatch_packet(uint8_t  dimensions,
                                                           const exe_symbol_kernel &kernel_symbol,
                                                           const hsa_signal &completion_signal,
                                                           const hsa_buffer &kernargs)
            : dimensions(dimensions), workgroup_size_x(0), workgroup_size_y(0), workgroup_size_z(0),
              grid_size_x(0), grid_size_y(0), grid_size_z(0),
              kernel_symbol(kernel_symbol), completion_signal(completion_signal), kernargs(kernargs) {
    }

    void hsa_kernel_dispatch_packet::set_workgroup_size(uint16_t x, uint16_t y, uint16_t z) {
        workgroup_size_x = x;
        workgroup_size_y = y;
        workgroup_size_z = z;
    }

    void hsa_kernel_dispatch_packet::set_grid_size(uint32_t x, uint32_t y, uint32_t z) {
        grid_size_x = x;
        grid_size_y = y;
        grid_size_z = z;
    }
}