//
// Created by koerberm on 06.03.18.
//

#ifndef EPHSA_HSA_PLATFORM_PROGRAMS_H
#define EPHSA_HSA_PLATFORM_PROGRAMS_H

#include <hsa/hsa.h>
#include <hsa/hsa_ext_finalize.h>
#include <vector>
#include <iostream>
#include <memory>
#include <cstring>


// No finalization in ROCM 2.0
//#define HSA_FINALIZER_VERSION_MAJOR 1
//#define HSA_FINALIZER_VERSION_MINOR 0
//typedef hsa_ext_finalizer_1_00_pfn_t hsa_finalizer_table_t;

namespace hsa {

    class hsa_agent;
    class hsa_platform;
//    class hsa_module;
    class hsa_executable;
    class hsa_code_object;
    class hsa_signal;
    class hsa_buffer;

//    No finalization in ROCm 2.x
//    std::unique_ptr<hsa_module> module_from_file(const char* file_name);
    std::shared_ptr<hsa_code_object> code_object_from_file(const std::string &filename, const hsa::hsa_agent &agent);

    /*
     * Code Object
     */

    class co_symbol {
    public:
        explicit co_symbol(hsa_code_symbol_t symbol);
        std::string name;
        std::string module_name;
        hsa_symbol_linkage_t linkage;
        bool is_definition;
    private:
        std::string extract_name(hsa_code_symbol_info_t len_prop, hsa_code_symbol_info_t val_prop);
    protected:
        const hsa_code_symbol_t handle;
    };

    class co_symbol_variable : public co_symbol {
    public:
        explicit co_symbol_variable(hsa_code_symbol_t symbol);
        hsa_variable_allocation_t allocation;
        hsa_variable_segment_t segment;
        uint32_t alignment;
        uint32_t size;
        bool is_const;

        friend std::ostream &operator<<(std::ostream &s, const co_symbol_variable &a);
    };

    class co_symbol_kernel : public co_symbol {
    public:
        explicit co_symbol_kernel(hsa_code_symbol_t symbol);
        uint32_t kernarg_segment_size;
        uint32_t kernarg_segment_alignment;
        uint32_t group_segment_size;
        uint32_t private_segment_size;
        bool dynamic_callstack;

        friend std::ostream &operator<<(std::ostream &s, const co_symbol_kernel &a);
    };

    class co_symbol_indirect_function : public co_symbol {
    public:
        explicit co_symbol_indirect_function(hsa_code_symbol_t symbol);
        uint32_t call_convention;

        friend std::ostream &operator<<(std::ostream &s, const co_symbol_indirect_function &a);
    };

    class hsa_code_object {
        friend class hsa_program;
        friend class hsa_executable;
    public:
        hsa_code_object(void *data, size_t size);
        ~hsa_code_object();
        hsa_code_object() = delete;
        hsa_code_object(const hsa_code_object &) = delete;
        const hsa_code_object &operator=(const hsa_code_object &) = delete;
    private:
        void* data;
        const size_t data_size;
        hsa_code_object_reader_t reader;
    };

//    No finalization in ROCM 2.0
//    /*
//     * Program
//     */
//
//    class hsa_module {
//        friend class hsa_program;
//    public:
//        explicit hsa_module(hsa_ext_module_t blob);
//        ~hsa_module();
//    private:
//        hsa_ext_module_t blob;
//    };
//
//
//    class hsa_program {
//    public:
//        hsa_machine_model_t machine_model;
//        hsa_profile_t profile;
//        hsa_default_float_rounding_mode_t rounding_mode;
//
//        explicit hsa_program(hsa_ext_program_t program);
//        ~hsa_program();
//
//        void add_module(std::shared_ptr<hsa_module> module);
//        std::unique_ptr<hsa_code_object> finalize(const hsa_agent &agent);
//    private:
//        hsa_ext_program_t handle;
//        hsa_finalizer_table_t table;
//        std::vector<std::shared_ptr<hsa_module>> modules;
//    };
//
//    class hsa_finalizer {
//    public:
//        hsa_finalizer();
//
//        std::unique_ptr<hsa_program> create_program(const hsa_agent &agent) const;
//        std::unique_ptr<hsa_program> create_program(const hsa_machine_model_t &machine_model,
//                                                    const hsa_profile_t &profile,
//                                                    const hsa_default_float_rounding_mode_t &rounding_mode) const;
//    private:
//        hsa_finalizer_table_t table;
//    };

    /*
     * Executable
     */
    class exe_symbol {
    public:
        explicit exe_symbol(hsa_executable_symbol_t symbol);
        std::string name;
        std::string module_name;
        hsa_symbol_linkage_t linkage;
        bool is_definition;
        hsa_agent_t agent;

    protected:
        const hsa_executable_symbol_t handle;

    private:
        std::string extract_name(hsa_executable_symbol_info_t len_prop, hsa_executable_symbol_info_t val_prop);
    };

    class exe_symbol_variable : public exe_symbol {
    public:
        uint64_t address;
        hsa_variable_allocation_t allocation;
        hsa_variable_segment_t segment;
        uint32_t alignment;
        uint32_t size;
        bool is_const;

        explicit exe_symbol_variable(hsa_executable_symbol_t symbol);

        friend std::ostream &operator<<(std::ostream &s, const exe_symbol_variable &a);
    };

    class exe_symbol_kernel : public exe_symbol {
    public:
        uint64_t kernel_object;
        uint32_t kernarg_segment_size;
        uint32_t kernarg_segment_alignment;
        uint32_t group_segment_size;
        uint32_t private_segment_size;
        bool dynamic_callstack;

        explicit exe_symbol_kernel(hsa_executable_symbol_t symbol);

        friend std::ostream &operator<<(std::ostream &s, const exe_symbol_kernel &a);
    };

    class exe_symbol_indirect_function : public exe_symbol {
    public:
        uint64_t function_object;
        uint32_t call_convention;

        explicit exe_symbol_indirect_function(hsa_executable_symbol_t symbol);

        friend std::ostream &operator<<(std::ostream &s, const exe_symbol_indirect_function &a);
    };




    class hsa_executable {
    public:
        hsa_profile_t profile;

        hsa_executable(hsa_profile_t profile = HSA_PROFILE_FULL, hsa_default_float_rounding_mode_t rounding_mode = HSA_DEFAULT_FLOAT_ROUNDING_MODE_DEFAULT);
        ~hsa_executable();

        hsa_executable(const hsa_executable&) = delete;
        const hsa_executable& operator=(const hsa_executable&) = delete;

        void define_global_variable( const std::string &name, void *address);
        void define_readonly_variable( const std::string &name, void *address, const hsa_agent &agent);
        void define_agent_global_variable( const std::string &name, void *address, const hsa_agent &agent);

        void load_code_object( std::shared_ptr<hsa_code_object> co, const hsa_agent &agent );

        void freeze();
        bool is_frozen() const { return frozen; };

        exe_symbol_kernel kernel_by_name(const std::string &name) const;

        const std::vector<exe_symbol_variable>& get_variable_symbols() const;

        const std::vector<exe_symbol_kernel>& get_kernel_symbols() const;

        const std::vector<exe_symbol_indirect_function>& get_indirect_function_symbols() const;

        friend std::ostream &operator<<(std::ostream &s, const hsa_executable &a);

    private:
        hsa_executable_t handle;
        bool frozen;
        std::vector<std::shared_ptr<hsa_code_object>> code_objects;
        std::vector<exe_symbol_variable> variable_symbols;
        std::vector<exe_symbol_kernel> kernel_symbols;
        std::vector<exe_symbol_indirect_function> indirect_function_symbols;

        template <class T, hsa_symbol_kind_t KIND>
        static hsa_status_t symbol_callback(hsa_executable_t exe, hsa_executable_symbol_t symbol, void *data);
    };

    class hsa_kernel_dispatch_packet {
    public:
        uint8_t  dimensions;
        uint16_t workgroup_size_x;
        uint16_t workgroup_size_y;
        uint16_t workgroup_size_z;
        uint32_t grid_size_x;
        uint32_t grid_size_y;
        uint32_t grid_size_z;
        const exe_symbol_kernel &kernel_symbol;
        const hsa_signal &completion_signal;
        const hsa_buffer &kernargs;

        hsa_kernel_dispatch_packet(uint8_t  dimensions, const exe_symbol_kernel &kernel_symbol, const hsa_signal &completion_signal, const hsa_buffer &kernargs);
        void set_workgroup_size( uint16_t x, uint16_t y = 1, uint16_t z = 1 );
        void set_grid_size( uint32_t x, uint32_t y = 1, uint32_t z = 1 );
    };
}


#endif //EPHSA_HSA_PLATFORM_PROGRAMS_H
