//
// Created by koerberm on 06.03.18.
//

#include "hsa_platform_util.h"

namespace hsa {
    /*
     * hsa_exception
     */

    hsa_exception::hsa_exception(hsa_status_t code) {
        const char *err;
        hsa_status_t stat = hsa_status_string(code, &err);
        if ( HSA_STATUS_SUCCESS == stat )
            msg = err;
        else
            msg = "Could not retrieve error string";
    }

    hsa_exception::hsa_exception(const hsa_exception&e) noexcept : msg(e.msg) {

    }

    hsa_exception::hsa_exception(const std::string &message) : msg(message) {
    }

    const char *hsa_exception::what() const noexcept {
        return msg.c_str();
    }

    std::ostream &operator<<(std::ostream &s, const hsa::hsa_exception &ex) {
        s << ex.msg;
        return s;
    }

    const char *map_machine_model(const hsa_machine_model_t &m) {
        switch (m) {
            case HSA_MACHINE_MODEL_LARGE: return "large";
            case HSA_MACHINE_MODEL_SMALL: return "small";
            default: throw hsa_exception("Unknown machine model.");
        }
    }

    const char *map_endianness(const hsa_endianness_t &e) {
        switch ( e ) {
            case HSA_ENDIANNESS_BIG: return "big";
            case HSA_ENDIANNESS_LITTLE: return "little";
            default: throw hsa_exception("Unknown endianness.");
        }
    }

    const char *map_device_type(const hsa_device_type_t &device) {
        switch ( device ) {
            case HSA_DEVICE_TYPE_CPU: return "CPU";
            case HSA_DEVICE_TYPE_GPU: return "GPU";
            case HSA_DEVICE_TYPE_DSP: return "DSP";
            default: throw hsa_exception("Unknown device type.");
        }
    }

    const char *map_rounding_mode(const hsa_default_float_rounding_mode_t &mode) {
        switch ( mode ) {
            case HSA_DEFAULT_FLOAT_ROUNDING_MODE_DEFAULT: return "default";
            case HSA_DEFAULT_FLOAT_ROUNDING_MODE_ZERO: return "zero";
            case HSA_DEFAULT_FLOAT_ROUNDING_MODE_NEAR: return "near";
            default: throw hsa_exception("Unknown rounding mode.");
        }
    }

    const char *map_agent_feature(const hsa_agent_feature_t &f) {
        switch (f) {
            case HSA_AGENT_FEATURE_KERNEL_DISPATCH: return "kernel dispatch";
            case HSA_AGENT_FEATURE_AGENT_DISPATCH: return "agent dispatch";
            default: return "none";
        }
    }

    const char *map_profile(const hsa_profile_t &p) {
        switch (p) {
            case HSA_PROFILE_BASE: return "base";
            case HSA_PROFILE_FULL: return "full";
            default: throw hsa_exception("Unknown HSA profile.");
        }
    }

    const char *map_queue_type(const hsa_queue_type_t &q) {
        switch (q) {
            case HSA_QUEUE_TYPE_SINGLE: return "single producer";
            case HSA_QUEUE_TYPE_MULTI: return "multi producer";
            default: throw hsa_exception("Unknown queue type.");
        }
    }

    const char *map_code_object_type(const hsa_code_object_type_t &c) {
        switch (c) {
            case HSA_CODE_OBJECT_TYPE_PROGRAM: return "program";
            default: throw hsa_exception("Unknown code object type.");
        }
    }

    const char *map_memory_segment(const hsa_region_segment_t &s) {
        switch ( s ) {
            case HSA_REGION_SEGMENT_GLOBAL: return "global";
            case HSA_REGION_SEGMENT_READONLY: return "readonly";
            case HSA_REGION_SEGMENT_PRIVATE: return "private";
            case HSA_REGION_SEGMENT_GROUP: return "group";
            default: throw hsa_exception("Unknown region segment.");
        }
    }
}
