//
// Created by koerberm on 06.03.18.
//

#ifndef EPHSA_HSA_PLATFORM_UTIL_H
#define EPHSA_HSA_PLATFORM_UTIL_H

#include <string>
#include <iostream>
#include <hsa/hsa.h>

namespace hsa {

    class hsa_exception : public std::exception {
    public:
        explicit hsa_exception(hsa_status_t code);
        explicit hsa_exception(const std::string& message);
        hsa_exception(const hsa_exception& e) noexcept;
        friend std::ostream& operator<<(std::ostream& s, const hsa_exception& ex);

        ~hsa_exception() override = default;

        const char *what() const noexcept override;
    private:
        std::string msg;
    };

    /*
     * Utilities
     */
    const char *map_machine_model(const hsa_machine_model_t &m);
    const char *map_endianness(const hsa_endianness_t &e);
    const char *map_device_type(const hsa_device_type_t &device);
    const char *map_rounding_mode(const hsa_default_float_rounding_mode_t &mode);
    const char *map_agent_feature(const hsa_agent_feature_t &f);
    const char *map_profile(const hsa_profile_t &p);
    const char *map_queue_type(const hsa_queue_type_t &q);
    const char *map_code_object_type(const hsa_code_object_type_t &c);
    const char *map_memory_segment(const  hsa_region_segment_t &s);
}

#endif //EPHSA_HSA_PLATFORM_UTIL_H