//
// Created by koerberm on 24.05.18.
//
#include "hsa_platform/hsa_platform.h"

void lanes_test(const hsa::hsa_platform &platform, const hsa::hsa_agent &agent, const hsa::hsa_executable &exe) {
    auto completion_signal = platform.create_signal(0);
    auto queue = agent.create_queue(agent.queue_max_size,HSA_QUEUE_TYPE_SINGLE,UINT32_MAX,UINT32_MAX);

    uint32_t group_size = agent.wave_front_size;

    uint32_t out1[64];
    uint32_t out2[64];

    memset(out1, 0, 64* sizeof(uint32_t) );
    memset(out2, 0, 64* sizeof(uint32_t) );


    // Args
    struct args_t {
        uint32_t *out1;
        uint32_t *out2;
    } args;

    args.out1 = out1;
    args.out2 = out2;


    auto kernel_symbol = exe.kernel_by_name("lane_kernel");
    auto kernarg_region = agent.get_kernarg_region();
    auto kernargs = kernarg_region.allocate(kernel_symbol.kernarg_segment_size,kernel_symbol.kernarg_segment_alignment);

    memcpy(kernargs->address<char>(), &args, sizeof(args));


    hsa::hsa_kernel_dispatch_packet p(1,kernel_symbol,*completion_signal,*kernargs);
    p.set_grid_size(group_size);
    p.set_workgroup_size(group_size);
    completion_signal->store_relaxed(1);
    queue->dispatch_kernel_packet(p);
    completion_signal->wait_relaxed(HSA_SIGNAL_CONDITION_LT, 1);

    std::cout << "Out1: [ ";
    for ( int i = 0; i < 64; i++ ) {
        std::cout << out1[i] << ", ";
    }
    std::cout << " ]" << std::endl;

    std::cout << "Out2: [ ";
    for ( int i = 0; i < 64; i++ ) {
        std::cout << out2[i] << ", ";
    }
    std::cout << " ]" << std::endl;

}

int main(int argc, char **argv) {
    (void) argc;
    (void) argv;

    hsa::hsa_platform platform;

    for ( auto &agent : platform.get_agents() ) {
        if (agent.features != HSA_AGENT_FEATURE_KERNEL_DISPATCH)
            continue;

        std::cout << "Using agent " << agent.name << std::endl;

        hsa::hsa_executable exe;

        auto co = hsa::code_object_from_file("src/rocm/cl/out/lanes.co", agent);

        exe.load_code_object(co, agent);
        exe.freeze();

        std::cout << "HSA Executable:" << std::endl;
        std::cout << exe << std::endl;

        lanes_test(platform, agent, exe);
        std::cout << "==========================================================" << std::endl;
    }

    return 0;
}