#include "../../shared.h"

__kernel void hells_join(
        int        probe_right,
        __global   void       *event,
        __constant window_def *window,
        __global   HELLS_JOIN_BITFIELD_TYPE       *results)
{
    const size_t global_id           = get_global_linear_id();
    const uint   elements_per_thread = sizeof(HELLS_JOIN_BITFIELD_TYPE) << 3;
    HELLS_JOIN_BITFIELD_TYPE res = 0;

    const ulong w_size = min(window->size, window->num_inserts);

    if ( probe_right ) {
        const T1 l = *((T1*) event);
        const T2 *r = window->base;

        for ( uint bit = 0; bit < elements_per_thread; bit++ ) {
            const size_t idx = global_id*elements_per_thread + bit;
            if ( idx < w_size && join_predicate(&l, &r[idx]) ) {
                res |= (1l << bit);
            }
        }
    }
    else {
        const T1 *l = window->base;
        const T2  r = *((T2*) event);

        for ( uint bit = 0; bit < elements_per_thread; bit++ ) {
            const size_t idx = global_id*elements_per_thread + bit;
            if ( idx < w_size && join_predicate(&l[idx], &r) ) {
                res |= (1l << bit);
            }
        }
    }
    results[global_id] = res;
}