//
// Created by koerberm on 06.03.19.
//

#ifndef EPHSA_HSA_COORDINATION_H
#define EPHSA_HSA_COORDINATION_H

#include "shared.h"
#include "cl_common.h"
#include "hsa_xlane.h"
#include "hsa_signals.h"

void hsa_coord_wait(bool signal_switch, const hsa_signal_t sig1, const hsa_signal_t sig2, hsa_signal_value_t num_groups) {
    // This is only called by a kernel with one work-item. No need to broadcast stuff
    const hsa_signal_t sig = signal_switch ? sig1 : sig2;
    hsa_signal_wait(sig, HSA_SIGNAL_CONDITION_EQ, num_groups, QUEUE_TERM_IDX, HSA_MEMORY_ORDER_SEQ_CST);
//    barrier(0);
}

void hsa_coord_fire( bool signal_switch, const hsa_signal_t sig1, const hsa_signal_t sig2 ) {
    // Tell processors to begin
    const hsa_signal_t sig = signal_switch ? sig1 : sig2;
    hsa_signal_store(sig, 0, HSA_MEMORY_ORDER_SEQ_CST);
}


bool hsa_group_wait( bool signal_switch, const hsa_signal_t sig1, const hsa_signal_t sig2, __local hsa_signal_value_t *bcast ) {
    hsa_signal_value_t wait_value;

    // Workgroup smaller than one wave
    if ( get_local_linear_size() <= WAVE_SIZE) {
        if ( lane_id() == 0 ) {
            const hsa_signal_t sig = signal_switch ? sig1 : sig2;
            wait_value = hsa_signal_wait(sig, HSA_SIGNAL_CONDITION_EQ, 0, QUEUE_TERM_IDX, HSA_MEMORY_ORDER_SEQ_CST);
        }
        //barrier(0);
        wait_value = broadcast_u64(wait_value,0);
    }
    // Larger groups
    else {
        if ( get_local_linear_id() == 0 ) {
            const hsa_signal_t sig = signal_switch ? sig1 : sig2;
            bcast[0] = hsa_signal_wait(sig, HSA_SIGNAL_CONDITION_EQ, 0, QUEUE_TERM_IDX, HSA_MEMORY_ORDER_SEQ_CST);
        }
        barrier(0);
        wait_value = bcast[0];
    }
    return wait_value == QUEUE_TERM_IDX;
}

void hsa_group_complete( bool signal_switch, const hsa_signal_t sig1, const hsa_signal_t sig2 ) {
    if ( get_local_linear_size() > WAVE_SIZE )
        barrier(0);
    // Only first thread of group signals completion
    if ( get_local_linear_id() == 0 ) {
        const hsa_signal_t sig = signal_switch ? sig2 : sig1;
        hsa_signal_add(sig, 1, HSA_MEMORY_ORDER_SEQ_CST);
    }
}

void hsa_group_ready(const hsa_signal_t sig1, const hsa_signal_t sig2) {
    hsa_group_complete(false,sig1,sig2);
}

#endif //EPHSA_HSA_COORDINATION_H
