//
// Created by koerberm on 06.03.19.
//

#ifndef EPHSA_HSA_EVENT_QUEUE_H
#define EPHSA_HSA_EVENT_QUEUE_H

#include "shared.h"
#include "hsa_xlane.h"
#include "hsa_signals.h"

typedef struct {
    hsa_signal_value_t read_idx;
    hsa_signal_value_t write_idx;
} queue_info;

hsa_signal_value_t get_queue_read_idx( const queue_def *q ) {
    return hsa_signal_load( q->r_signal, HSA_MEMORY_ORDER_SEQ_CST );
}

hsa_signal_value_t get_queue_write_idx( const queue_def *q ) {
    return hsa_signal_load( q->w_signal, HSA_MEMORY_ORDER_SEQ_CST );
}

void add_to_queue_read_idx( const queue_def *q, hsa_signal_value_t value) {
    return hsa_signal_add( q->r_signal, value, HSA_MEMORY_ORDER_SEQ_CST );
}

void add_to_queue_write_idx( const queue_def *q, hsa_signal_value_t value ) {
    return hsa_signal_add( q->w_signal, value, HSA_MEMORY_ORDER_SEQ_CST );
}

hsa_signal_value_t update_queue_write_idx( const queue_def *q, hsa_signal_value_t expected_value, hsa_signal_value_t new_value ) {
    hsa_signal_value_t result = new_value;

    while ( result != expected_value && result != QUEUE_TERM_IDX )
        result = __ockl_hsa_signal_cas( q->w_signal,expected_value,new_value, HSA_MEMORY_ORDER_SEQ_CST );

    return result;
}

void set_queue_write_idx( const queue_def *q, hsa_signal_value_t new_value ) {
    __ockl_hsa_signal_store( q->w_signal,new_value, HSA_MEMORY_ORDER_SEQ_CST );
}

/*void wait_for_queue_elements( const queue_def *q, hsa_signal_value_t num_elems, queue_info *result ) {
    result->read_idx = get_queue_read_idx(q);
    const hsa_signal_value_t min_write_idx = get_queue_read_idx(q) + num_elems;
    result->write_idx = hsa_signal_wait(q->w_signal, HSA_SIGNAL_CONDITION_GTE, min_write_idx, QUEUE_TERM_IDX, HSA_MEMORY_ORDER_SEQ_CST);
}*/

void* get_queue_element( const queue_def *q, const hsa_signal_value_t read_idx, const size_t element_size ) {
    const long physical_idx = read_idx & (long) q->mask;
    return ( q->base + physical_idx*element_size );
}

#endif //EPHSA_HSA_EVENT_QUEUE_H
