//
// Created by koerberm on 06.03.19.
//

#ifndef EPHSA_HSA_SIGNALS_H
#define EPHSA_HSA_SIGNALS_H

#include "/opt/rocm/include/ockl_hsa.h"

#define HSA_MEMORY_ORDER_RELAXED __ockl_memory_order_relaxed
#define HSA_MEMORY_ORDER_ACQUIRE __ockl_memory_order_acquire
#define HSA_MEMORY_ORDER_RELEASE __ockl_memory_order_release
#define HSA_MEMORY_ORDER_SEQ_CST __ockl_memory_order_seq_cst
#define HSA_MEMORY_ORDER_ACQ_REL __ockl_memory_order_acq_rel

typedef __ockl_memory_order hsa_memory_order;

hsa_signal_value_t hsa_signal_load(const hsa_signal_t sig, hsa_memory_order memory_order) {
    return __ockl_hsa_signal_load(sig,memory_order);
}

void hsa_signal_add(const hsa_signal_t sig, hsa_signal_value_t val, hsa_memory_order memory_order) {
    return __ockl_hsa_signal_add(sig,val,memory_order);
}

void hsa_signal_store(const hsa_signal_t sig, hsa_signal_value_t val, hsa_memory_order memory_order) {
    return __ockl_hsa_signal_store(sig,val,memory_order);
}

hsa_signal_value_t hsa_signal_wait(const hsa_signal_t sig, const hsa_signal_condition_t condition, const hsa_signal_value_t cmp,
                                   const hsa_signal_value_t break_value, hsa_memory_order memory_order ) {

    do {
        hsa_signal_value_t value = __ockl_hsa_signal_load(sig, memory_order);
        if ( condition == HSA_SIGNAL_CONDITION_GTE && value >= cmp )
            return value;
        else if ( condition == HSA_SIGNAL_CONDITION_EQ && value == cmp )
            return value;
        else if ( condition == HSA_SIGNAL_CONDITION_NE && value != cmp )
            return value;
        else if ( condition == HSA_SIGNAL_CONDITION_LT && value < cmp )
            return value;
        else if ( value == break_value )
            return value;
    } while (true);
}

#endif //EPHSA_HSA_SIGNALS_H
