//
// Created by koerberm on 30.11.18.
//

#include "ep/event.h"
#include "ep/event_queue.h"
#include "ep/cpu_operators.h"
#include "ep/hsa_operators.h"
#include "ep/dispatch_operators.h"
#include "ep/cpu_parallel_operators.h"

int main(int argc, char **argv) {
    (void) argc;
    (void) argv;

    hsa::hsa_platform platform;
    for ( auto &agent : platform.get_agents() ) {
        if (agent.features != HSA_AGENT_FEATURE_KERNEL_DISPATCH)
            continue;

        std::cout << "Using agent " << agent.name << std::endl;

        hsa::hsa_executable exe;

        std::vector<std::string> cos = {
                "src/rocm/cl/out/hsa_filter_normal.co",
                "src/rocm/cl/out/hsa_filter_wf_normal.co",
                "src/rocm/cl/out/hells_filter_normal.co"
        };

        for (auto &file : cos) {
            try {
                auto co = hsa::code_object_from_file(file, agent);
                exe.load_code_object(co, agent);
            } catch (const hsa::hsa_exception &e) {
                std::cout << "Error loading module (" << file << "): " << e.what() << std::endl;
                return 1;
            }
        }

        exe.freeze();


        auto qin = std::make_shared<ep::event_queue<event>>(64, platform, agent);
        auto qout = std::make_shared<ep::event_queue<event>>(64, platform, agent);

        ep::parallel_filter<event> f(
                qin,
                qout,
                1, 64, 2,
                [](const event &e) { return filter_predicate(&e); },
                platform, agent
        );

        ep::event_printer<event> p(
                qout,
                1, 64
        );

        qin->add({1, {0.3f, 0, 0, 0, 0, 0}});
        qin->add({1, {1.2f, 0, 0, 0, 0, 0}});
        qin->add({1, {0.5f, 0, 0, 0, 0, 0}});

        qin->wait_for_space(64, 1000000000);

        qin->add({2, {0.3f, 0, 0, 0, 0, 0}});
        qin->add({2, {1.2f, 0, 0, 0, 0, 0}});
        qin->add({2, {0.5f, 0, 0, 0, 0, 0}});

        qin->wait_for_space(64, 1000000000);
        qout->wait_for_space(64, 1000000000);
        std::this_thread::sleep_for(std::chrono::milliseconds(100));
    }
}