//
// Created by koerberm on 08.06.18.
//

#include <iostream>
#include <fstream>
#include "cl_util.h"


namespace cl_util {

    void get_gpu_devices(std::vector<cl::Device> &devices) {
        ;
        std::vector<cl::Platform> platform;
        cl::Platform::get(&platform);

        cl::Context context;
        cl::Device device;
        for(auto &p : platform) {
            std::vector<cl::Device> pldev;

            p.getDevices(CL_DEVICE_TYPE_GPU, &pldev);


            for(auto &d : pldev) {
                if (!d.getInfo<CL_DEVICE_AVAILABLE>()) continue;
                devices.push_back(d);
            }
        }
    }


    cl::Program program_from_file(const std::string &file, cl::Context ctx) {
        std::string src;

        std::ifstream in(file, std::ios::in);
        if (!in)
            throw cl::Error(CL_BUILD_PROGRAM_FAILURE, "Could not open source file.");

        in.seekg(0, std::ios::end);
        long t_res = in.tellg();
        if ( t_res < 0 )
            throw cl::Error(CL_BUILD_PROGRAM_FAILURE, "Could not determine length of file.");

        src.resize(static_cast<size_t >(t_res));
        in.seekg(0, std::ios::beg);
        src.assign((std::istreambuf_iterator<char>(in)),
                   std::istreambuf_iterator<char>());
        in.close();


        return cl::Program(ctx, cl::Program::Sources(
                1, std::make_pair(src.c_str(), src.length())
        ));
    }

    void build_program(cl::Program &prog, const cl::Device &device) {
        try {
            prog.build({device});
        } catch (const cl::Error &err) {
            std::cerr << "OpenCL compilation error" << std::endl
                      << prog.getBuildInfo<CL_PROGRAM_BUILD_LOG>(device)
                      << std::endl;
            throw;
        }
    }
}
